-- Copyright 2012 Alberto Sáez
--
--
-- This file is part of Witchaven the Roguelike.
--
--    Witchaven the Roguelike is free software: you can redistribute it and/or modify
--    it under the terms of the GNU General Public License as published by
--    the Free Software Foundation, either version 3 of the License, or
--    (at your option) any later version.
--
--    Witchaven the Roguelike is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--    GNU General Public License for more details.
--
--    You should have received a copy of the GNU General Public License
--    along with Witchaven the Roguelike.  If not, see <http://www.gnu.org/licenses/>.

m = {}

map = {}
traps = {}
materials = {}
mapWidth = 55
mapHeight = 45
rooms = {}
doors = {}
corridors = {}
viewMap = {} -- 0 not visited, 1 = viewing, -1 = fog of war

map.floorASC = "·"
map.wallASC = "#"

function m.draw()
	love.graphics.setFont(myFont)
	for i = 1, mapHeight do
		for j = 1, mapWidth do
			if viewMap[i][j] == 1 then viewMap[i][j] = -1 end
		end
	end
	
	for i = player.x - playerLOS, player.x + playerLOS do
		for j = player.y - playerLOS, player.y + playerLOS do
			local A = los(player.x, player.y, i, j, function(x,y)
				if map[y][x] == '#' then 
					viewMap[y][x] = 1
					return false
				end
				viewMap[y][x] = 1
				return true
			end)
		end
	end
	
	enemy.awake()
	
	local wr1, wg1, wb1 = math.random(40, 50), math.random(75, 100), math.random(200,255)
	local wr2, wg2, wb2 = math.random(10, 15), math.random(20, 35), math.random(110, 150)
	local lr1, lg1, lb1 = math.random(200, 255), math.random(150, 180), math.random(30, 50)
	local lr2, lg2, lb2 =  math.random(110, 150), math.random(40, 60), math.random(20, 35)
	
	for i = 1, mapHeight do
		for j = 1, mapWidth do
			if viewMap[i][j] == 0 then 
				if showLevelMap == false then
					color(0,0,0,255)
				else
					color(70,0,0,255)
				end
			elseif viewMap[i][j] == 1 then
				if map[i][j] == "·" then 
					color(floorColorLOS)
				elseif map[i][j] == "%" or map[i][j] == "," or map[i][j] == ":" or map[i][j] == "\"" or map[i][j] == "^" then
					color(255, 50, 25, 255)
				elseif map[i][j] == "#" then
					color(wallColorLOS)
				elseif map[i][j] == "<" or map[i][j] == ">" then
					color({200, 200, 200, 255})
				elseif map[i][j] == "~" then
					color({wr1, wg1, wb1, 255})
				elseif map[i][j] == ";" then
					color({lr1, lg1, lb1, 255})
				else
					color(decorColorLOS)
				end
			elseif viewMap[i][j] == -1 then
				if map[i][j] == "·" then
					color(floorColorFOW)
				elseif map[i][j] == "#" then
					color(wallColorFOW)
				elseif map[i][j] == "<" or map[i][j] == ">" then
					color({75, 75, 75, 255})
				elseif map[i][j] == "~" then
					color({wr2, wg2, wb2, 255})
				elseif map[i][j] == ";" then
					color({lr2, lg2, lb2, 255})
				else
					color(decorColorFOW)
				end
			end
			
			if map[i][j] == "·" or map[i][j] == "#" or  map[i][j] == "<" or map[i][j] == ">" then
				love.graphics.setFont(myFont)
				lprint(map[i][j], (fontSize - 5)*j, (fontSize - 5)*i)
			else
				love.graphics.setFont(myFont2)
				lprint(map[i][j], (fontSize - 5)*j, (fontSize - 5)*i + 2)
			end
		end
	end
end

function m.generate(maxRooms, minRoomSize, maxRoomSize)
	local numOfRooms = 0
	local freeSpace = true
	
	if maxRoomSize == nil then maxRoomSize = 10 end
	if minRoomSize == nil then minRoomSize = 3 end
	
	for i = 1, mapHeight do
		map[i] = {}
		viewMap[i] = {}
		materials[i] = {}
		for j = 1, mapWidth do
			map[i][j] = "#"
			viewMap[i][j] = 0
			materials[i][j] = 0
		end
	end
	
	while numOfRooms < maxRooms do
		local x, y, w, h
		x = math.random(3, mapWidth - 3)
		y = math.random(3, mapHeight - 3)
		
		w = math.random(minRoomSize, maxRoomSize)
		h = math.random(minRoomSize, maxRoomSize)
		
		if x + w <= mapWidth - 2 and y + h <= mapHeight - 2 then
			for i = y - 1, y+h + 1 do
				for j = x-1, x+w+1 do
					if map[i][j] == "·" then
						freeSpace = false
					end
				end
			end
			
			if freeSpace == true then
				table.insert(rooms, {x = x, y = y, w = w, h = h, door = false})
				for i,v in ipairs(rooms) do
					for j = v.x, v.x + v.w do
						for k = v.y, v.y + v.h do
							map[k][j] = "·"
						end
					end
				end
				numOfRooms = numOfRooms + 1
			end
		end
		freeSpace = true
	end
	
	for i,v in ipairs(rooms) do
		if math.random(0, 100) >=50 then
			--door in north/south wall
			x = math.random(v.x, v.x + v.w)
			
			if math.random(0, 100) >= 50 then
				map[v.y-1][x] = "#"
				table.insert(doors,{x = x, y = v.y-1, connect = false})
			else
				map[v.y + v.h + 1][x] = "#"
				table.insert(doors,{x = x, y = v.y + v.h + 1, connect = false})
			end
		else
			--door in east/west wall
			y = math.random(v.y, v.y + v.h)
			
			if math.random(0, 100) >= 50 then
				map[y][v.x - 1] = "#"
				table.insert(doors,{x = v.x - 1, y = y, connect = false})
			else
				map[y][v.x + v.w + 1] = "#"
				table.insert(doors,{x = v.x + v.w + 1, y = y, connect = false})
			end
		end
	end
	
	local jumper = require("jumper")
	pather = jumper(map, "#")
	pather:setMode("ORTHOGONAL")
	
	for i,v in ipairs(doors) do
		aDoor = i
		
		while aDoor == i do
			aDoor = math.random(1, #(doors))
		end
		
		--if doors[aDoor].connect == false then
			myPath = pather:getPath(v.x, v.y, doors[aDoor].x, doors[aDoor].y)
			
			if myPath ~= nil then
				for i, v in ipairs(myPath) do
					table.insert(corridors, {x = v.x, y = v.y})
				end
				
				doors[aDoor].connect = true
			end
		--end
	end
	
	-- lava and water rooms
	for i = 1, 1 + player.depth/3 do
		local poolType = math.random(1, 100)
		-- water
		if poolType < 50 then
			randomRoom = math.random(   1, #(rooms)    )
			for i = rooms[randomRoom].x, rooms[randomRoom].x + rooms[randomRoom].w do
				for j = rooms[randomRoom].y, rooms[randomRoom].y + rooms[randomRoom].h do
					if math.random(1, 100) <= 90 then
						map[j][i] = "~"
					end
				end
			end
		elseif poolType >= 50 then
			randomRoom = math.random(   1, #(rooms)    )
			for i = rooms[randomRoom].x, rooms[randomRoom].x + rooms[randomRoom].w do
				for j = rooms[randomRoom].y, rooms[randomRoom].y + rooms[randomRoom].h do
					if math.random(1, 100) <= 90 then
						map[j][i] = ";"
					end
				end
			end
		end
	end
	
	
	for i,v in ipairs(doors) do
		map[v.y][v.x] = "·"
	end
	
	for i,v in ipairs(corridors) do
		map[v.y][v.x] = "·"
	end
	
	rooms = nil
	doors = nil
	corridors = nil
	
	while true do
		x = math.random (1, mapWidth)
		y = math.random (1, mapHeight)
		
		if map[y][x] == "·" then
			player.x = x
			player.y = y
			if player.depth == 1 then map[y][x] = "<" end
			break
		end
	end
	
	if player.depth < 15 then
		while true do
			x = math.random (1, mapWidth)
			y = math.random (1, mapHeight)
			
			if map[y][x] == "·" then
				map[y][x] = ">"
				break
			end
		end
	end
	
	for i = 1, mapHeight do
		traps[i] = {}
		for j = 1, mapWidth do
			traps[i][j] = map[i][j]
		end
	end
	
	for i = 1, mapWidth do
		for j = 1, mapHeight do
		
			if traps[j][i] == "·" and math.random(1, 250 - player.depth*3.33) <= 1 then
				traps[j][i] = "^"
			end
		end
	end
	
	enemy.recalculateBlockMap()
end

function m.generateDecor()
	local shrineExists = nil
	for i = 1, mapWidth do
		for j = 1, mapHeight do
			if math.random(1, 100) <= 1 and map[j][i] == "·" then
				local decor = math.random(1, 100)
				
				if decor <= 15 then
					map[j][i] = "±"
				elseif decor > 15 and decor <= 30 then
					map[j][i] = "¥"
				elseif decor > 30 and decor <= 50 then
					map[j][i] = "Ø"
				elseif decor > 50 and decor <= 65 then
					map[j][i] = "¤"
				elseif decor > 65 and decor <= 80 then
					map[j][i] = "÷"
				elseif decor > 80 then
					if shrineExists ~= true then
						map[j][i] = "_"
						shrineExists = true
					end
				end
			end
		end
	end
end

function m.generateItems(chanceToSpawn, numOfTries)
	local totalTries = 0
	
	for i = 1, mapWidth do
		for j = 1, mapHeight do
			if map[j][i] == "·" or map[j][i] == ";" or map[j][i] == "~" then
				if math.random(1, 500) <= chanceToSpawn and totalTries < numOfTries then
					totalTries = totalTries + 1
					
					aux = m.generateRandomItem(i, j)
				end
			end
		end
	end
	
	-- Teleporting pentagram, dlvls 1 - 14
	if player.depth < 15 then
		while true do
			local x, y = math.random(1, mapWidth), math.random(1, mapHeight)
			
			if map[y][x] == "·" then
				inventory.addItemToMap("&", {255,240,190, 255}, x, y, "t", true, 0, 0, 0, 0)
				break
			end
		end
	end
end

function m.spawnBlood(x, y)
	local bloodyRemains = "."
	local remains = math.random(1,3)
	
	if remains == 1 then
		bloodyRemains = ","
	elseif remains == 2 then
		bloodyRemains = ":"
	elseif remains == 3 then
		bloodyRemains = "\""
	end
	
	map[y][x] = bloodyRemains
end

function m.generateRandomItem(x, y)
	local thisItem = math.random(1, 185)
					
	if thisItem <= 30 then
		-- weapon
		local thisWeapon = math.random(1, 9)
		inventory.addItemToMap(")", {150, 150, 150, 255}, x, y, "w", false, thisWeapon, weapons[thisWeapon].bdmg, math.random(0, 3), math.random(60, 100))
		
	elseif thisItem > 30 and thisItem <= 50 then
		-- potion
		local thisPotion = math.random(1, 5)
		inventory.addItemToMap("!", potions[thisPotion].color, x, y, "p", false, thisPotion, 0, 0, 0)
		
	elseif thisItem > 50 and thisItem <= 60 then
		-- Arrow
		inventory.addItemToMap("/", {235, 180, 80, 255}, x, y, "r", true, 0, 0, 0, 0)
		
	elseif thisItem > 60 and thisItem <= 90 then
		-- gold
		inventory.addItemToMap("$", {255, 255, 0, 255}, x, y, "g", true, math.random(player.depth, player.depth + player.depth/2), 0, 0, 0)
		
	elseif thisItem > 90 and thisItem <= 100 then
		-- scroll
		inventory.addItemToMap("?", {200, 120, 25, 255}, x, y, "s", true, math.random(1, 8), 0, 0, 0)
		
	elseif thisItem > 100 and thisItem <= 120 then
		-- armor
		inventory.addItemToMap("[", {150, 150, 150, 255}, x, y, "a", true, math.random(2, 4)*10, math.random(75, 100), 0, 0)
		
	elseif thisItem > 120 and thisItem <= 130 then
		-- ank
		inventory.addItemToMap("£", {205, 205, 150, 255}, x, y, "ank", true, 0, 0, 0, 0)
	
	elseif thisItem > 130 and thisItem <= 140 then
		-- Amulet Mist
		inventory.addItemToMap("¢", {150, 150, 180, 255}, x, y, "mist", true, 0, 0, 0, 0)
		
	elseif thisItem > 140 and thisItem <= 145 then
		-- Shadow Amulet
		inventory.addItemToMap("§", {135, 135, 135, 255}, x, y, "shadow", true, 0, 0, 0, 0)
		
	elseif thisItem > 145 and thisItem <= 150 then
		-- Crystal Staff
		inventory.addItemToMap("|", {150, 200, 150, 255}, x, y, "cstaff", true, 0, 0, 0, 0)
		
	elseif thisItem > 150 and thisItem <= 160 then
		-- Helmet
		inventory.addItemToMap("µ", {200, 200, 200, 255}, x, y, "helmet", true, 0, 0, 0, 0)
	
	elseif thisItem > 160 and thisItem <= 170 then
		-- Horn
		inventory.addItemToMap("¬", {200,25,125, 255}, x, y, "horn", true, 0, 0, 0, 0)						
		
	elseif thisItem > 170 and thisItem <= 185 then
		-- jewel
		inventory.addItemToMap("*", {255, 255, 255, 255}, x, y, "j", true, math.random(1, #(shardNames)), 0, 0, 0)
	end
	
	return mapDescriptions[mapItems[#mapItems].char]
end

function m.generateEnemies()
	local numOfEnemies = math.random(1, player.depth) + player.depth/2
	local totalEnemies = 0
	
	while totalEnemies < numOfEnemies do
		x = math.random(1, mapWidth)
		y = math.random(1, mapHeight)
		
		if map[y][x] ~= "#" and x ~= player.x and y ~= player.y then
			if #(enemyList) <= 0 then
				if player.depth == 15 then
					enemy.addPrototype(22, x, y)
					totalEnemies = totalEnemies + 1
				else
					enemy.addPrototype(math.random(1, 21), x, y)
					totalEnemies = totalEnemies + 1
				end
			end
			
			for i = 1, #(enemyList) do
				if enemyList[i].x ~= x and enemyList[i].y ~= y then
					enemy.addPrototype(math.random(1, 21), x, y)
					totalEnemies = totalEnemies + 1
					break
				end
			end
		end
	end
end

function m.applyItem()
	if map[player.y][player.x] == ">" then
		if player.hasPentagram == true then
			player.depth = player.depth + 1
			game.restart(true)
			logMessage("You enter the map teleporting rune.", {100, 255, 100, 255})
			game.turn()
		else
			logMessage("You need to find the Teleporting Pentagram to activate the teleport.", {100, 255, 100, 255})
		end
	elseif map[player.y][player.x] == "<" then
		logMessage("It's now too late for going back to the surface!!!")
	
	
	elseif map[player.y][player.x] == "Ø" then
		local randomEffect = math.random(1, 100)
		
		if randomEffect <= 30 then
			aux = m.generateRandomItem(player.x, player.y)
			logMessage("There is an " .. aux .. " in the barrel!!!")
		elseif randomEffect > 30 and randomEffect <= 65 then
			player.health = player.health + 5
			if player.health > player.maxHealth then player.health = player.maxHealth end
			logMessage("Aaahh some fresh water refreshes you... (+5 HP)")
		elseif randomEffect > 65 and randomEffect <= 100 then
			logMessage("The " .. mapDescriptions[map[player.y][player.x]] .. " only contains some rubbish and decaying pieces of flesh")
		end
		
		map[player.y][player.x] = "·"
		
		game.turn()
		
	elseif map[player.y][player.x] == "¤" then
		local randomEffect = math.random(1, 100)
		
		if randomEffect <= 45 then
			aux = m.generateRandomItem(player.x, player.y)
			logMessage("There is an " .. aux .. " in the chest!!!")
		elseif randomEffect > 45 and randomEffect <= 60 then
			player.inflictDamage(20, true)
			logMessage("It was a trapped chest!!!", {255, 0, 0, 255})
		elseif randomEffect > 60 and randomEffect <= 100 then
			logMessage("The chest was empty...")
		end
		
		map[player.y][player.x] = "·"
		
		game.turn()
		
	elseif map[player.y][player.x] == "¥" then
		if math.random(1, 100) <= 30 then
			if playerWeapons[player.selectedWeapon].durability >= 100 then
				logMessage("You guess why are you trying to repair a full repaired weapon...")
			else
				playerWeapons[player.selectedWeapon].durability = playerWeapons[player.selectedWeapon].durability  + 2
				logMessage("Using the intense flames, you repair a bit your weapon")
			end
		elseif math.random(1, 100) >= 75 then
			playerWeapons[player.selectedWeapon].durability = playerWeapons[player.selectedWeapon].durability  - 2
			if playerWeapons[player.selectedWeapon].durability < 0 then
				playerWeapons[player.selectedWeapon].durability  = 0
			end
			logMessage("You try to repair your weapon with the intense flames, but you only make it worst")
		else
			logMessage("Ouch.. it burns!!", {255, 200, 0, 255})
			player.health = player.health - 3
		end
		
		game.turn()
	elseif map[player.y][player.x] == "_" then
		if player.usedEnchanting == false then
			player.shrineMod = 200
			player.usedEnchanting = true
			logMessage("Your weapons start to shine with a cyan glow", {0, 255, 200, 255})
		else
			logMessage("It seems that this shrine doesn't have any magic", {0, 255, 200, 255})
		end
	else
		logMessage("There is nothing here you can use")
	end
end

function m.save()
	local saveMap = nil
	
	saveMap = "mapHeight = " .. mapHeight
	saveMap = saveMap .. "\nmapWidth = " .. mapWidth
	saveMap = saveMap .. "\nmap = " .. ttostring(map)
	saveMap = saveMap .. "\nviewMap = " .. ttostring(viewMap)
	saveMap = saveMap .. "\ntraps = " .. ttostring(traps)
	
	success = love.filesystem.write("game.map", saveMap)
	
	return success
end

function m.load()
	ok, chunk = pcall(love.filesystem.load, "game.map" )
	
	return ok, chunk
end
